@@ -2,10 +2,11 @@
 
 use pyo3::exceptions::{PyAttributeError, PyIndexError, PyValueError};
 use pyo3::types::{PyDict, PyList, PyMapping, PySequence, PySlice, PyType};
-use pyo3::{prelude::*, py_run, PyCell};
-use std::{isize, iter};
+use pyo3::{prelude::*, py_run};
+use std::iter;
+use std::sync::Mutex;
 
-mod common;
+mod test_utils;
 
 #[pyclass]
 struct EmptyClass;
@@ -14,22 +15,22 @@ struct EmptyClass;
 struct ExampleClass {
     #[pyo3(get, set)]
     value: i32,
-    _custom_attr: Option<i32>,
+    custom_attr: Option<i32>,
 }
 
 #[pymethods]
 impl ExampleClass {
-    fn __getattr__(&self, py: Python, attr: &str) -> PyResult<PyObject> {
+    fn __getattr__(&self, py: Python<'_>, attr: &str) -> PyResult<Py<PyAny>> {
         if attr == "special_custom_attr" {
-            Ok(self._custom_attr.into_py(py))
+            Ok(self.custom_attr.into_pyobject(py)?.into_any().unbind())
         } else {
             Err(PyAttributeError::new_err(attr.to_string()))
         }
     }
 
-    fn __setattr__(&mut self, attr: &str, value: &PyAny) -> PyResult<()> {
+    fn __setattr__(&mut self, attr: &str, value: &Bound<'_, PyAny>) -> PyResult<()> {
         if attr == "special_custom_attr" {
-            self._custom_attr = Some(value.extract()?);
+            self.custom_attr = Some(value.extract()?);
             Ok(())
         } else {
             Err(PyAttributeError::new_err(attr.to_string()))
@@ -38,7 +39,7 @@ impl ExampleClass {
 
     fn __delattr__(&mut self, attr: &str) -> PyResult<()> {
         if attr == "special_custom_attr" {
-            self._custom_attr = None;
+            self.custom_attr = None;
             Ok(())
         } else {
             Err(PyAttributeError::new_err(attr.to_string()))
@@ -63,21 +64,20 @@ impl ExampleClass {
     }
 }
 
-fn make_example(py: Python) -> &PyCell<ExampleClass> {
-    Py::new(
+fn make_example(py: Python<'_>) -> Bound<'_, ExampleClass> {
+    Bound::new(
         py,
         ExampleClass {
             value: 5,
-            _custom_attr: Some(20),
+            custom_attr: Some(20),
         },
     )
     .unwrap()
-    .into_ref(py)
 }
 
 #[test]
 fn test_getattr() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
         assert_eq!(
             example_py
@@ -104,7 +104,7 @@ fn test_getattr() {
 
 #[test]
 fn test_setattr() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
         example_py.setattr("special_custom_attr", 15).unwrap();
         assert_eq!(
@@ -120,7 +120,7 @@ fn test_setattr() {
 
 #[test]
 fn test_delattr() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
         example_py.delattr("special_custom_attr").unwrap();
         assert!(example_py.getattr("special_custom_attr").unwrap().is_none());
@@ -129,26 +129,23 @@ fn test_delattr() {
 
 #[test]
 fn test_str() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
-        assert_eq!(example_py.str().unwrap().to_str().unwrap(), "5");
+        assert_eq!(example_py.str().unwrap(), "5");
     })
 }
 
 #[test]
 fn test_repr() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
-        assert_eq!(
-            example_py.repr().unwrap().to_str().unwrap(),
-            "ExampleClass(value=5)"
-        );
+        assert_eq!(example_py.repr().unwrap(), "ExampleClass(value=5)");
     })
 }
 
 #[test]
 fn test_hash() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
         assert_eq!(example_py.hash().unwrap(), 5);
     })
@@ -156,11 +153,11 @@ fn test_hash() {
 
 #[test]
 fn test_bool() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let example_py = make_example(py);
-        assert!(example_py.is_true().unwrap());
+        assert!(example_py.is_truthy().unwrap());
         example_py.borrow_mut().value = 0;
-        assert!(!example_py.is_true().unwrap());
+        assert!(!example_py.is_truthy().unwrap());
     })
 }
 
@@ -176,7 +173,7 @@ impl LenOverflow {
 
 #[test]
 fn len_overflow() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let inst = Py::new(py, LenOverflow).unwrap();
         py_expect_exception!(py, inst, "len(inst)", PyOverflowError);
     });
@@ -189,27 +186,29 @@ pub struct Mapping {
 
 #[pymethods]
 impl Mapping {
-    fn __len__(&self, py: Python) -> usize {
-        self.values.as_ref(py).len()
+    fn __len__(&self, py: Python<'_>) -> usize {
+        self.values.bind(py).len()
     }
 
-    fn __getitem__<'a>(&'a self, key: &'a PyAny) -> PyResult<&'a PyAny> {
-        let any: &PyAny = self.values.as_ref(key.py()).as_ref();
+    fn __getitem__<'py>(&self, key: &Bound<'py, PyAny>) -> PyResult<Bound<'py, PyAny>> {
+        let any: &Bound<'py, PyAny> = self.values.bind(key.py());
         any.get_item(key)
     }
 
-    fn __setitem__(&self, key: &PyAny, value: &PyAny) -> PyResult<()> {
-        self.values.as_ref(key.py()).set_item(key, value)
+    fn __setitem__<'py>(&self, key: &Bound<'py, PyAny>, value: &Bound<'py, PyAny>) -> PyResult<()> {
+        self.values.bind(key.py()).set_item(key, value)
     }
 
-    fn __delitem__(&self, key: &PyAny) -> PyResult<()> {
-        self.values.as_ref(key.py()).del_item(key)
+    fn __delitem__(&self, key: &Bound<'_, PyAny>) -> PyResult<()> {
+        self.values.bind(key.py()).del_item(key)
     }
 }
 
 #[test]
 fn mapping() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
+        PyMapping::register::<Mapping>(py).unwrap();
+
         let inst = Py::new(
             py,
             Mapping {
@@ -218,8 +217,7 @@ fn mapping() {
         )
         .unwrap();
 
-        //
-        let mapping: &PyMapping = inst.as_ref(py).downcast().unwrap();
+        let mapping: &Bound<'_, PyMapping> = inst.bind(py).cast().unwrap();
 
         py_assert!(py, inst, "len(inst) == 0");
 
@@ -246,14 +244,14 @@ fn mapping() {
 }
 
 #[derive(FromPyObject)]
-enum SequenceIndex<'a> {
+enum SequenceIndex<'py> {
     Integer(isize),
-    Slice(&'a PySlice),
+    Slice(Bound<'py, PySlice>),
 }
 
 #[pyclass]
 pub struct Sequence {
-    values: Vec<PyObject>,
+    values: Vec<Py<PyAny>>,
 }
 
 #[pymethods]
@@ -262,13 +260,13 @@ impl Sequence {
         self.values.len()
     }
 
-    fn __getitem__(&self, index: SequenceIndex) -> PyResult<PyObject> {
+    fn __getitem__(&self, index: SequenceIndex<'_>, py: Python<'_>) -> PyResult<Py<PyAny>> {
         match index {
             SequenceIndex::Integer(index) => {
                 let uindex = self.usize_index(index)?;
                 self.values
                     .get(uindex)
-                    .map(Clone::clone)
+                    .map(|o| o.clone_ref(py))
                     .ok_or_else(|| PyIndexError::new_err(index))
             }
             // Just to prove that slicing can be implemented
@@ -276,7 +274,7 @@ impl Sequence {
         }
     }
 
-    fn __setitem__(&mut self, index: isize, value: PyObject) -> PyResult<()> {
+    fn __setitem__(&mut self, index: isize, value: Py<PyAny>) -> PyResult<()> {
         let uindex = self.usize_index(index)?;
         self.values
             .get_mut(uindex)
@@ -294,7 +292,7 @@ impl Sequence {
         }
     }
 
-    fn append(&mut self, value: PyObject) {
+    fn append(&mut self, value: Py<PyAny>) {
         self.values.push(value);
     }
 }
@@ -316,10 +314,12 @@ impl Sequence {
 
 #[test]
 fn sequence() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
+        PySequence::register::<Sequence>(py).unwrap();
+
         let inst = Py::new(py, Sequence { values: vec![] }).unwrap();
 
-        let sequence: &PySequence = inst.as_ref(py).downcast().unwrap();
+        let sequence: &Bound<'_, PySequence> = inst.bind(py).cast().unwrap();
 
         py_assert!(py, inst, "len(inst) == 0");
 
@@ -346,49 +346,48 @@ fn sequence() {
         // indices.
         assert!(sequence.len().is_err());
         // however regular python len() works thanks to mp_len slot
-        assert_eq!(inst.as_ref(py).len().unwrap(), 0);
+        assert_eq!(inst.bind(py).len().unwrap(), 0);
 
         py_run!(py, inst, "inst.append(0)");
         sequence.set_item(0, 5).unwrap();
-        assert_eq!(inst.as_ref(py).len().unwrap(), 1);
+        assert_eq!(inst.bind(py).len().unwrap(), 1);
 
         assert_eq!(sequence.get_item(0).unwrap().extract::<u8>().unwrap(), 5);
         sequence.del_item(0).unwrap();
 
-        assert_eq!(inst.as_ref(py).len().unwrap(), 0);
+        assert_eq!(inst.bind(py).len().unwrap(), 0);
     });
 }
 
 #[pyclass]
 struct Iterator {
-    iter: Box<dyn iter::Iterator<Item = i32> + Send>,
+    iter: Mutex<Box<dyn iter::Iterator<Item = i32> + Send>>,
 }
 
 #[pymethods]
 impl Iterator {
-    fn __iter__(slf: PyRef<Self>) -> PyRef<Self> {
+    fn __iter__(slf: PyRef<'_, Self>) -> PyRef<'_, Self> {
         slf
     }
 
-    fn __next__(mut slf: PyRefMut<Self>) -> Option<i32> {
-        slf.iter.next()
+    fn __next__(slf: PyRefMut<'_, Self>) -> Option<i32> {
+        slf.iter.lock().unwrap().next()
     }
 }
 
 #[test]
 fn iterator() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-
-    let inst = Py::new(
-        py,
-        Iterator {
-            iter: Box::new(5..8),
-        },
-    )
-    .unwrap();
-    py_assert!(py, inst, "iter(inst) is inst");
-    py_assert!(py, inst, "list(inst) == [5, 6, 7]");
+    Python::attach(|py| {
+        let inst = Py::new(
+            py,
+            Iterator {
+                iter: Mutex::new(Box::new(5..8)),
+            },
+        )
+        .unwrap();
+        py_assert!(py, inst, "iter(inst) is inst");
+        py_assert!(py, inst, "list(inst) == [5, 6, 7]");
+    });
 }
 
 #[pyclass]
@@ -406,41 +405,14 @@ struct NotCallable;
 
 #[test]
 fn callable() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-
-    let c = Py::new(py, Callable).unwrap();
-    py_assert!(py, c, "callable(c)");
-    py_assert!(py, c, "c(7) == 42");
-
-    let nc = Py::new(py, NotCallable).unwrap();
-    py_assert!(py, nc, "not callable(nc)");
-}
-
-#[allow(deprecated)]
-mod deprecated {
-    use super::*;
-
-    #[pyclass]
-    struct Callable;
-
-    #[pymethods]
-    impl Callable {
-        #[__call__]
-        fn __call__(&self, arg: i32) -> i32 {
-            arg * 6
-        }
-    }
-
-    #[test]
-    fn callable() {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
+    Python::attach(|py| {
         let c = Py::new(py, Callable).unwrap();
         py_assert!(py, c, "callable(c)");
         py_assert!(py, c, "c(7) == 42");
-    }
+
+        let nc = Py::new(py, NotCallable).unwrap();
+        py_assert!(py, nc, "not callable(nc)");
+    });
 }
 
 #[pyclass]
@@ -460,17 +432,16 @@ impl SetItem {
 
 #[test]
 fn setitem() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-
-    let c = PyCell::new(py, SetItem { key: 0, val: 0 }).unwrap();
-    py_run!(py, c, "c[1] = 2");
-    {
-        let c = c.borrow();
-        assert_eq!(c.key, 1);
-        assert_eq!(c.val, 2);
-    }
-    py_expect_exception!(py, c, "del c[1]", PyNotImplementedError);
+    Python::attach(|py| {
+        let c = Bound::new(py, SetItem { key: 0, val: 0 }).unwrap();
+        py_run!(py, c, "c[1] = 2");
+        {
+            let c = c.borrow();
+            assert_eq!(c.key, 1);
+            assert_eq!(c.val, 2);
+        }
+        py_expect_exception!(py, c, "del c[1]", PyNotImplementedError);
+    });
 }
 
 #[pyclass]
@@ -487,16 +458,15 @@ impl DelItem {
 
 #[test]
 fn delitem() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-
-    let c = PyCell::new(py, DelItem { key: 0 }).unwrap();
-    py_run!(py, c, "del c[1]");
-    {
-        let c = c.borrow();
-        assert_eq!(c.key, 1);
-    }
-    py_expect_exception!(py, c, "c[1] = 2", PyNotImplementedError);
+    Python::attach(|py| {
+        let c = Bound::new(py, DelItem { key: 0 }).unwrap();
+        py_run!(py, c, "del c[1]");
+        {
+            let c = c.borrow();
+            assert_eq!(c.key, 1);
+        }
+        py_expect_exception!(py, c, "c[1] = 2", PyNotImplementedError);
+    });
 }
 
 #[pyclass]
@@ -517,18 +487,17 @@ impl SetDelItem {
 
 #[test]
 fn setdelitem() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-
-    let c = PyCell::new(py, SetDelItem { val: None }).unwrap();
-    py_run!(py, c, "c[1] = 2");
-    {
+    Python::attach(|py| {
+        let c = Bound::new(py, SetDelItem { val: None }).unwrap();
+        py_run!(py, c, "c[1] = 2");
+        {
+            let c = c.borrow();
+            assert_eq!(c.val, Some(2));
+        }
+        py_run!(py, c, "del c[1]");
         let c = c.borrow();
-        assert_eq!(c.val, Some(2));
-    }
-    py_run!(py, c, "del c[1]");
-    let c = c.borrow();
-    assert_eq!(c.val, None);
+        assert_eq!(c.val, None);
+    });
 }
 
 #[pyclass]
@@ -543,13 +512,12 @@ impl Contains {
 
 #[test]
 fn contains() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-
-    let c = Py::new(py, Contains {}).unwrap();
-    py_run!(py, c, "assert 1 in c");
-    py_run!(py, c, "assert -1 not in c");
-    py_expect_exception!(py, c, "assert 'wrong type' not in c", PyTypeError);
+    Python::attach(|py| {
+        let c = Py::new(py, Contains {}).unwrap();
+        py_run!(py, c, "assert 1 in c");
+        py_run!(py, c, "assert -1 not in c");
+        py_expect_exception!(py, c, "assert 'wrong type' not in c", PyTypeError);
+    });
 }
 
 #[pyclass]
@@ -557,8 +525,8 @@ struct GetItem {}
 
 #[pymethods]
 impl GetItem {
-    fn __getitem__(&self, idx: &PyAny) -> PyResult<&'static str> {
-        if let Ok(slice) = idx.cast_as::<PySlice>() {
+    fn __getitem__(&self, idx: &Bound<'_, PyAny>) -> PyResult<&'static str> {
+        if let Ok(slice) = idx.cast::<PySlice>() {
             let indices = slice.indices(1000)?;
             if indices.start == 100 && indices.stop == 200 && indices.step == 1 {
                 return Ok("slice");
@@ -574,13 +542,12 @@ impl GetItem {
 
 #[test]
 fn test_getitem() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
+    Python::attach(|py| {
+        let ob = Py::new(py, GetItem {}).unwrap();
 
-    let ob = Py::new(py, GetItem {}).unwrap();
-
-    py_assert!(py, ob, "ob[1] == 'int'");
-    py_assert!(py, ob, "ob[100:200:1] == 'slice'");
+        py_assert!(py, ob, "ob[1] == 'int'");
+        py_assert!(py, ob, "ob[100:200:1] == 'slice'");
+    });
 }
 
 #[pyclass]
@@ -598,11 +565,11 @@ impl ClassWithGetAttr {
 
 #[test]
 fn getattr_doesnt_override_member() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-    let inst = PyCell::new(py, ClassWithGetAttr { data: 4 }).unwrap();
-    py_assert!(py, inst, "inst.data == 4");
-    py_assert!(py, inst, "inst.a == 8");
+    Python::attach(|py| {
+        let inst = Py::new(py, ClassWithGetAttr { data: 4 }).unwrap();
+        py_assert!(py, inst, "inst.data == 4");
+        py_assert!(py, inst, "inst.a == 8");
+    });
 }
 
 #[pyclass]
@@ -613,18 +580,24 @@ struct ClassWithGetAttribute {
 
 #[pymethods]
 impl ClassWithGetAttribute {
-    fn __getattribute__(&self, _name: &str) -> u32 {
-        self.data * 2
+    fn __getattribute__(&self, name: &str) -> PyResult<u32> {
+        if name == "data" {
+            Ok(self.data * 2)
+        } else {
+            Err(PyAttributeError::new_err(
+                "this message will be swallowed by default `__getattr__`",
+            ))
+        }
     }
 }
 
 #[test]
 fn getattribute_overrides_member() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-    let inst = PyCell::new(py, ClassWithGetAttribute { data: 4 }).unwrap();
-    py_assert!(py, inst, "inst.data == 8");
-    py_assert!(py, inst, "inst.y == 8");
+    Python::attach(|py| {
+        let inst = Py::new(py, ClassWithGetAttribute { data: 4 }).unwrap();
+        py_assert!(py, inst, "inst.data == 8");
+        py_expect_exception!(py, inst, "inst.y == 8", PyAttributeError, "y");
+    });
 }
 
 #[pyclass]
@@ -653,44 +626,44 @@ impl ClassWithGetAttrAndGetAttribute {
 
 #[test]
 fn getattr_and_getattribute() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-    let inst = PyCell::new(py, ClassWithGetAttrAndGetAttribute).unwrap();
-    py_assert!(py, inst, "inst.exists == 42");
-    py_assert!(py, inst, "inst.lucky == 57");
-    py_expect_exception!(py, inst, "inst.error", PyValueError);
-    py_expect_exception!(py, inst, "inst.unlucky", PyAttributeError);
+    Python::attach(|py| {
+        let inst = Py::new(py, ClassWithGetAttrAndGetAttribute).unwrap();
+        py_assert!(py, inst, "inst.exists == 42");
+        py_assert!(py, inst, "inst.lucky == 57");
+        py_expect_exception!(py, inst, "inst.error", PyValueError);
+        py_expect_exception!(py, inst, "inst.unlucky", PyAttributeError);
+    });
 }
 
 /// Wraps a Python future and yield it once.
 #[pyclass]
 #[derive(Debug)]
 struct OnceFuture {
-    future: PyObject,
+    future: Py<PyAny>,
     polled: bool,
 }
 
 #[pymethods]
 impl OnceFuture {
     #[new]
-    fn new(future: PyObject) -> Self {
+    fn new(future: Py<PyAny>) -> Self {
         OnceFuture {
             future,
             polled: false,
         }
     }
 
-    fn __await__(slf: PyRef<Self>) -> PyRef<Self> {
+    fn __await__(slf: PyRef<'_, Self>) -> PyRef<'_, Self> {
         slf
     }
 
-    fn __iter__(slf: PyRef<Self>) -> PyRef<Self> {
+    fn __iter__(slf: PyRef<'_, Self>) -> PyRef<'_, Self> {
         slf
     }
-    fn __next__(mut slf: PyRefMut<Self>) -> Option<PyObject> {
-        if !slf.polled {
-            slf.polled = true;
-            Some(slf.future.clone())
+    fn __next__<'py>(&mut self, py: Python<'py>) -> Option<&Bound<'py, PyAny>> {
+        if !self.polled {
+            self.polled = true;
+            Some(self.future.bind(py))
         } else {
             None
         }
@@ -698,11 +671,11 @@ impl OnceFuture {
 }
 
 #[test]
+#[cfg(not(target_arch = "wasm32"))] // Won't work without wasm32 event loop (e.g., Pyodide has WebLoop)
 fn test_await() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-    let once = py.get_type::<OnceFuture>();
-    let source = r#"
+    Python::attach(|py| {
+        let once = py.get_type::<OnceFuture>();
+        let source = cr#"
 import asyncio
 import sys
 
@@ -716,11 +689,12 @@ if sys.platform == "win32" and sys.version_info >= (3, 8, 0):
 
 asyncio.run(main())
 "#;
-    let globals = PyModule::import(py, "__main__").unwrap().dict();
-    globals.set_item("Once", once).unwrap();
-    py.run(source, Some(globals), None)
-        .map_err(|e| e.print(py))
-        .unwrap();
+        let globals = PyModule::import(py, "__main__").unwrap().dict();
+        globals.set_item("Once", once).unwrap();
+        py.run(source, Some(&globals), None)
+            .map_err(|e| e.display(py))
+            .unwrap();
+    });
 }
 
 #[pyclass]
@@ -737,7 +711,7 @@ impl AsyncIterator {
         }
     }
 
-    fn __aiter__(slf: PyRef<Self>) -> PyRef<Self> {
+    fn __aiter__(slf: PyRef<'_, Self>) -> PyRef<'_, Self> {
         slf
     }
 
@@ -747,11 +721,11 @@ impl AsyncIterator {
 }
 
 #[test]
+#[cfg(not(target_arch = "wasm32"))] // Won't work without wasm32 event loop (e.g., Pyodide has WebLoop)
 fn test_anext_aiter() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-    let once = py.get_type::<OnceFuture>();
-    let source = r#"
+    Python::attach(|py| {
+        let once = py.get_type::<OnceFuture>();
+        let source = cr#"
 import asyncio
 import sys
 
@@ -769,14 +743,15 @@ if sys.platform == "win32" and sys.version_info >= (3, 8, 0):
 
 asyncio.run(main())
 "#;
-    let globals = PyModule::import(py, "__main__").unwrap().dict();
-    globals.set_item("Once", once).unwrap();
-    globals
-        .set_item("AsyncIterator", py.get_type::<AsyncIterator>())
-        .unwrap();
-    py.run(source, Some(globals), None)
-        .map_err(|e| e.print(py))
-        .unwrap();
+        let globals = PyModule::import(py, "__main__").unwrap().dict();
+        globals.set_item("Once", once).unwrap();
+        globals
+            .set_item("AsyncIterator", py.get_type::<AsyncIterator>())
+            .unwrap();
+        py.run(source, Some(&globals), None)
+            .map_err(|e| e.display(py))
+            .unwrap();
+    });
 }
 
 /// Increment the count when `__get__` is called.
@@ -795,29 +770,28 @@ impl DescrCounter {
     /// Each access will increase the count
     fn __get__<'a>(
         mut slf: PyRefMut<'a, Self>,
-        _instance: &PyAny,
-        _owner: Option<&PyType>,
+        _instance: &Bound<'_, PyAny>,
+        _owner: Option<&Bound<'_, PyType>>,
     ) -> PyRefMut<'a, Self> {
         slf.count += 1;
         slf
     }
     /// Allow assigning a new counter to the descriptor, copying the count across
-    fn __set__(&self, _instance: &PyAny, new_value: &mut Self) {
+    fn __set__(&self, _instance: &Bound<'_, PyAny>, new_value: &mut Self) {
         new_value.count = self.count;
     }
     /// Delete to reset the counter
-    fn __delete__(&mut self, _instance: &PyAny) {
+    fn __delete__(&mut self, _instance: &Bound<'_, PyAny>) {
         self.count = 0;
     }
 }
 
 #[test]
 fn descr_getset() {
-    let gil = Python::acquire_gil();
-    let py = gil.python();
-    let counter = py.get_type::<DescrCounter>();
-    let source = pyo3::indoc::indoc!(
-        r#"
+    Python::attach(|py| {
+        let counter = py.get_type::<DescrCounter>();
+        let source = pyo3_ffi::c_str!(
+            r#"
 class Class:
     counter = Counter()
 
@@ -840,12 +814,13 @@ assert c.counter.count == 4
 del c.counter
 assert c.counter.count == 1
 "#
-    );
-    let globals = PyModule::import(py, "__main__").unwrap().dict();
-    globals.set_item("Counter", counter).unwrap();
-    py.run(source, Some(globals), None)
-        .map_err(|e| e.print(py))
-        .unwrap();
+        );
+        let globals = PyModule::import(py, "__main__").unwrap().dict();
+        globals.set_item("Counter", counter).unwrap();
+        py.run(source, Some(&globals), None)
+            .map_err(|e| e.display(py))
+            .unwrap();
+    });
 }
 
 #[pyclass]
@@ -854,14 +829,14 @@ struct NotHashable;
 #[pymethods]
 impl NotHashable {
     #[classattr]
-    const __hash__: Option<PyObject> = None;
+    const __hash__: Option<Py<PyAny>> = None;
 }
 
 #[test]
 fn test_hash_opt_out() {
     // By default Python provides a hash implementation, which can be disabled by setting __hash__
     // to None.
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let empty = Py::new(py, EmptyClass).unwrap();
         py_assert!(py, empty, "hash(empty) is not None");
 
@@ -876,10 +851,11 @@ struct DefaultedContains;
 
 #[pymethods]
 impl DefaultedContains {
-    fn __iter__(&self, py: Python) -> PyObject {
-        PyList::new(py, &["a", "b", "c"])
-            .as_ref()
-            .iter()
+    fn __iter__(&self, py: Python<'_>) -> Py<PyAny> {
+        PyList::new(py, ["a", "b", "c"])
+            .unwrap()
+            .as_any()
+            .try_iter()
             .unwrap()
             .into()
     }
@@ -890,10 +866,11 @@ struct NoContains;
 
 #[pymethods]
 impl NoContains {
-    fn __iter__(&self, py: Python) -> PyObject {
-        PyList::new(py, &["a", "b", "c"])
-            .as_ref()
-            .iter()
+    fn __iter__(&self, py: Python<'_>) -> Py<PyAny> {
+        PyList::new(py, ["a", "b", "c"])
+            .unwrap()
+            .as_any()
+            .try_iter()
             .unwrap()
             .into()
     }
@@ -901,14 +878,14 @@ impl NoContains {
     // Equivalent to the opt-out const form in NotHashable above, just more verbose, to confirm this
     // also works.
     #[classattr]
-    fn __contains__() -> Option<PyObject> {
+    fn __contains__() -> Option<Py<PyAny>> {
         None
     }
 }
 
 #[test]
 fn test_contains_opt_out() {
-    Python::with_gil(|py| {
+    Python::attach(|py| {
         let defaulted_contains = Py::new(py, DefaultedContains).unwrap();
         py_assert!(py, defaulted_contains, "'a' in defaulted_contains");
 
